@@ -1,29 +1,43 @@
-from subprocess import call
+# Copyright (C) 2022-2026  Hsin Yuan Yeh <iapyeh@gmail.com>
+#
+# This file is part of Sshscript.
+#
+# SSHScript is free software; you can redistribute it and/or modify it under the
+# terms of the MIT License.
+#
+# SSHScript is distributed in the hope that it will be useful, but WITHOUT ANY
+# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
+# A PARTICULAR PURPOSE.  See the MIT License for more details.
+#
+# You should have received a copy of the MIT License along with SSHScript;
+# if not, write to the Free Software Foundation, Inc.,
+# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.
+#
 import threading
 import paramiko
-import hashlib
 import asyncio
 import stat
-import traceback
 import time
 import os
+import sys
 import re
 import random
+import warnings
 import __main__
 from io import StringIO
 from logging import DEBUG
 try:
     from .sshscriptdollar import SSHScriptDollar
-    from .sshscripterror import SSHScriptError, SSHScriptExit, SSHScriptBreak, getLogger
+    from .sshscripterror import SSHScriptError, SSHScriptExit, SSHScriptBreak,SSHScriptCareful, getLogger
 except ImportError:
     from sshscriptdollar import SSHScriptDollar
-    from sshscripterror import SSHScriptError, SSHScriptExit, SSHScriptBreak, getLogger
+    from sshscripterror import SSHScriptError, SSHScriptExit, SSHScriptBreak,SSHScriptCareful, getLogger
 
 # replace @{var} in $shell-command
 pvar = re.compile('(\b?)\@\{(.+)\}')
 # replace $.stdout, $.stderr to _c.stdout, _c.stderr, $.host in $LINE
 pstd = re.compile('\$\.([a-z]+)')
-#$.stdout in f-string, but not {{$.stdout}} 
+#$.stdout in string, but not {{$.stdout}} 
 pstdInCurlyBracket = re.compile(r"([^{]?){((?:\\.|[^}\\])*)}([^}])",re.M) 
 # looking for @include( to self.open( in Py
 pAtInclude = re.compile('^( *?)\$.include\(([^\)]+)\)',re.M)
@@ -51,8 +65,10 @@ pCurlyBrackets = re.compile('^[ \t]*?\}')
 pqoute1 = re.compile('([fr]?)("{3})(.+?)"{3}',re.S)
 pqoute2 = re.compile("([fr]?)('{3})(.+?)'{3}",re.S)
 # ' .. ', "...", r'..', r"..""
-pqoute3 = re.compile("([fr]?)(')(.+?)'",re.M)
-pqoute4 = re.compile('([fr]?)(")(.+?)"',re.M)
+pqoute3 = re.compile("([fr]?)'.*?'",re.M)
+pqoute4 = re.compile('([fr]?)".*?"',re.M)
+pqoute5 = re.compile('#____(.+?)____#',re.M)
+
 # replace $.stdout, $.stderr to _c.stdout, _c.stderr
 # replace $.open(), $.subopen() to SSHScript.inContext.
 pQuoted = re.compile('\(.*\)')
@@ -77,21 +93,9 @@ def pstdSub(m):
 __main__.SSHScriptExportedNames = SSHScriptExportedNames
 __main__.SSHScriptExportedNamesByAlias = SSHScriptExportedNamesByAlias
 
-'''
-def pstdSub_origin(m):
-    pre,post = m.groups()
-    if post in SSHScriptExportedNames:
-        return f'{pre}SSHScript.getContext(1).{post}'
-    elif post in SSHScriptDollar.exportedProperties:
-        return f'{pre}_c.{post}'
-    elif post in SSHScriptExportedNamesByAlias:
-        return f'{pre}SSHScript.getContext(1).{SSHScriptExportedNamesByAlias[post]}'
-    #elif post in SSHScriptClsExportedFuncs:
-    #    return f'{pre}SSHScript.{post}'
-    else:
-        # SSHScript.inContext's property
-        return f'{pre}SSHScript.getContext(1).{post}'
-'''
+global newline, bNewline
+newline = os.linesep # \n
+bNewline = os.linesep.encode('utf8') # b'\n'
 # resolve f'''{$.stdout}''' 
 def pstdInCurlyBracketSub(m):
     f= f'{m.group(1)}{{{pstd.sub(pstdSub,m.group(2))}}}{m.group(3)}'
@@ -165,8 +169,11 @@ class SSHScript(object):
         t = threading.current_thread()
         cls.contextLocker.acquire()
         if obj is None:
-            del cls.context[t]
-            #getLogger().log(DEBUG,f'set context to {obj} in <Thread {t.ident}>')
+            try:
+                del cls.context[t]
+            except KeyError:
+                #getLogger().log(DEBUG,f'context of <Thread {t.ident}> is already None')
+                pass
         else:
             cls.context[t] = obj
             obj.logger.log(DEBUG,f'set context to {obj} in <Thread {t.ident}>')
@@ -201,6 +208,7 @@ class SSHScript(object):
     
     @classmethod
     def include(cls,prefix,abspath,alreadyIncluded=None):
+        global newline
         if alreadyIncluded is None:
             alreadyIncluded = {}
         
@@ -228,12 +236,12 @@ class SSHScript(object):
         
         # omit leading prefix
         rows = []
-        for line in content.split('\n'):
+        for line in content.splitlines():
             # skip "__export__ = [ "line
             if line.replace(' ','').startswith('__export__=['):
                 continue
             rows.append(prefix + line[prefixLen:])
-        script =  '\n'.join(rows)
+        script =  newline.join(rows)
         
         if pAtInclude.search(script):
             # do include again (nested @include)
@@ -301,14 +309,16 @@ class SSHScript(object):
                 SSHScript.setContext(self)
        
 
+        ## 這兩個可能有bug on threading
         # settings when exec command (see client.exec_command)
-        self._paranoid = parent._paranoid if parent else False
-
+        self._careful = parent._careful if parent else False
         self._timeout = parent._timeout if parent else None #blocking
         
         # 累積的行號(多檔案showScript時使用)
         # v1.1.13起若有--debug時，顯示個別檔案的行號，這樣比較適合debug使用
         self.lineNumberCount = 0
+
+        self.mute = os.environ.get('MUTE_WARNING')
     
     @property
     def host(self):
@@ -403,15 +413,16 @@ class SSHScript(object):
     def __exit__(self,*args):
         self.close() # close all subsession if any
 
-    def touchIO(self,mesg=None):
+    def touchIO(self):
         self.lock.acquire()
         self._lastIOTime = time.time()
         self.lock.release()
 
     async def waitio(self,waitingInterval):
-        remain = waitingInterval - (time.time() - self._lastIOTime)
-        if remain > 0:
-            await asyncio.sleep(remain)    
+        while True:
+            remain = waitingInterval - (time.time() - self._lastIOTime)
+            if remain <= 0: break
+            await asyncio.sleep(remain)
 
     def getSocketWithProxyCommand(self,argsOfProxyCommand):
         return paramiko.ProxyCommand(argsOfProxyCommand)        
@@ -521,9 +532,13 @@ class SSHScript(object):
     
     @export2Dollar
     def paranoid(self,yes):
-        self._paranoid = yes
-    
-    
+        warnings.warn(f'''paranoid() has renamed to careful()''',UserWarning,stacklevel=0)        
+        self._careful = yes
+
+    @export2Dollar
+    def careful(self,yes):
+        self._careful = yes
+
     @export2Dollar
     def close(self):
         
@@ -566,27 +581,10 @@ class SSHScript(object):
     @export2Dollar
     def pkey(self,pathOfRsaPrivate):
         # 往自己或往上找連線
-        '''
-        if self._client:
-            _,stdout,_ = self._client.exec_command(f'cat "{pathOfRsaPrivate}"')
-            keyfile = StringIO(stdout.read().decode('utf8'))
-            return paramiko.RSAKey.from_private_key(keyfile)
-        else:
-            connectedParentSession = self.connectedParent
-            if connectedParentSession:
-                _,stdout,_ = connectedParentSession._client.exec_command(f'cat "{pathOfRsaPrivate}"')
-                keyfile = StringIO(stdout.read().decode('utf8'))
-                return paramiko.RSAKey.from_private_key(keyfile)
-            else:
-                # localhost
-                with open(pathOfRsaPrivate) as fd:
-                    return paramiko.RSAKey.from_private_key(fd)
-        '''
-        client = self.client
-        if client:
+        if self.client:
             try:
                 self.lock.acquire()
-                _,stdout,_ = client.exec_command(f'cat "{pathOfRsaPrivate}"')
+                _,stdout,_ = self.client.exec_command(f'cat "{pathOfRsaPrivate}"')
                 exitcode = stdout.channel.recv_exit_status()
                 if exitcode > 0:
                     raise SSHScriptError(f'failed to get pkey from {pathOfRsaPrivate}, exitcode = {exitcode}')
@@ -615,8 +613,11 @@ class SSHScript(object):
             raise SSHScriptError(f'uploading src "{src}" must be a file',503)
         
         dst = os.path.normpath(dst)
-        if not dst.startswith('/'):
-            raise SSHScriptError(f'uploading dst "{dst}" must be absolute path',504)
+        if sys.platform == 'win32':
+            # recorrect back, since ftp accecpt /
+            dst = dst.replace('\\','/')        
+        if not dst.startswith('/') and not self.mute:
+            warnings.warn(f'''uploading dst "{dst}" is not absolute path''',UserWarning,stacklevel=0)
         
         self.log(DEBUG,f'upload {src} to {dst}')
 
@@ -640,7 +641,7 @@ class SSHScript(object):
             # (see unittest/c0.spy)
             if not dstbasename == srcbasename:
                 if not os.path.splitext(srcbasename)[1] == os.path.splitext(dstbasename)[1]:
-                    dst = os.path.join(dst,srcbasename)
+                    dst = f'{dst}/{srcbasename}'
 
             def checking(dst,foldersToMake):
                 dstDir = os.path.dirname(dst)
@@ -673,8 +674,8 @@ class SSHScript(object):
                 else:
                     # REF: https://stackoverflow.com/questions/18205731/how-to-check-a-remote-path-is-a-file-or-a-directory
                     if stat.S_ISDIR(dststat.st_mode):
-                        # is folder
-                        dst = os.path.join(dst,srcbasename)
+                        # is folder，不要用os.path.join,若win32會變成倒斜線
+                        dst = f'{dst}/{srcbasename}'
                         try:
                             dststat = self.sftp.stat(dst)
                         except FileNotFoundError:
@@ -696,17 +697,13 @@ class SSHScript(object):
         if dst is None:
             dst = os.getcwd()
 
-        # 先往下找(因為self.sftp已經會先往下找，因此這裡應該是不必要)<== 這個版本v1.1.13不會
-        #if self.subSession:
-        #    return self.subSession.downaload(src,dst)
+        #v1.1.13之後不必往下找self.subSession
 
-        if not src.startswith('/'):
-            raise SSHScriptError(f'downloading src "{src}" must be absolute path',505)
-        
-        if not dst.startswith('/'):
-            #raise SSHScriptError(f'downloading dst "{dst}" must be absolute path',506)
-            dst = os.path.abspath(dst)
+        if not src.startswith('/') and not self.mute:
+            warnings.warn(f'''downloading src "{src}" is not absolute path''',UserWarning,stacklevel=0)        
         
+        dst = os.path.abspath(dst)
+
         # if dst is a folder, append the filename of src to dst
         if os.path.isdir(dst):
             dst = os.path.join(dst,os.path.basename(src))
@@ -719,6 +716,7 @@ class SSHScript(object):
 
     def parseScript(self,script,initLine=None,_locals=None,quotes=None):
         # parse script at top-level session
+        global newline
         scriptPath = _locals.get('__file__') if _locals else __file__
         def pAtIncludeSub(m):
             prefix, path = m.groups()
@@ -728,20 +726,19 @@ class SSHScript(object):
         
         def pqouteSub12(m):
             fr,quote,content = m.groups()
-            #hash = hashlib.md5(content.encode('utf8'))
-            #key = f'#____{fr or "-"}{hash.hexdigest()}____#'
             key = f'#____{fr}_{hash(content)}{random.random()}____#'
-            quotes[key] = m.group(0)#(fr,quote,content)
+            quotes[key] = m.group(0)
             return key
         def pqouteSub34(m):
             # 只替換內容中有 $.stdout等特殊內容的字串
-            fr,quote,content = m.groups()
-            if not pstd.search(content):
-                return m.group(0)
-            #hash = hashlib.md5(content.encode('utf8'))
-            #key = f'#____{fr or "-"}{hash.hexdigest()}____#'
+            fr = m.group(1)
+            content = m.group(0)
+            #print(content, pstd.search(content), pqoute5.search(content))
+            if not (pstd.search(content) or pqoute5.search(content)):
+                return content
+            #print(content,'<<')
             key = f'#____{fr}_{hash(content)}{random.random()}____#'
-            quotes[key] = m.group(0)# (fr,quote,content)
+            quotes[key] = content
             return key
         # expend @include()
         script = pAtInclude.sub(pAtIncludeSub,script)
@@ -753,7 +750,7 @@ class SSHScript(object):
         script = pqoute4.sub(pqouteSub34,script)
         # replace with @open( to "with open("
         script = pAtFuncSWith.sub('\\1with SSHScript.getContext(1).\\3',script)
-        listOfLines = script.split('\n')
+        listOfLines = script.splitlines()
         # listOfLines: [string]
         lines = LineGenerator(listOfLines)
         # lines is a generator
@@ -778,7 +775,7 @@ class SSHScript(object):
                 if m:  break
                 cmd = pstd.sub(pstdSub,nextLine.lstrip())
                 rowsInCurlyBrackets.append(cmd)
-            return self.parseScript('\n'.join(rowsInCurlyBrackets),_locals=_locals,quotes=quotes)
+            return self.parseScript(newline.join(rowsInCurlyBrackets),_locals=_locals,quotes=quotes)
 
         while True:
             try:
@@ -861,7 +858,7 @@ class SSHScript(object):
                                 continue
                             asPartOfWith = m.group(1) # the " as ..." part
                             break
-                        shellCommandLines.extend(self.parseScript('\n'.join(rowsInCurlyBrackets),_locals=_locals,quotes=quotes))
+                        shellCommandLines.extend(self.parseScript(newline.join(rowsInCurlyBrackets),_locals=_locals,quotes=quotes))
                 
                 #
                 # 逐行處理開始
@@ -894,12 +891,12 @@ class SSHScript(object):
                             except SSHScriptError as e:
                                 if e.code == 403:
                                     print('--- dump start ---')
-                                    print('\n'.join(lines.lines[c:lines.cursor + 1]))
+                                    print(newline.join(lines.lines[c:lines.cursor + 1]))
                                     print('--- dump end ---')
                                     raise e
                             inlineScript.extend(shellCommandLines)
                         
-                        cmd = '\n'.join(inlineScript)
+                        cmd = newline.join(inlineScript)
                         # 分成先建立SSHScriptDollar物件,再執行的2步驟。先取名為 __c 而不是 _c ，
                         # 此__c執行之後，再指定成為 _c 
                         # 因為命令中可能會呼叫 _c（例如 $.stdout)，係指前次執行後的_c,這樣才不會混淆。
@@ -907,12 +904,22 @@ class SSHScript(object):
                         # 這樣可以在SSHScriptError丟出之前把_c設定為__c
                         indentOfTryBlock = '    '
                         rows.append(f'{indent}try:')
+                        print('cmd=',cmd)
                         if cmd.startswith('#____'):
+                            #____r, #____f,#____
+                            fr = 1 if cmd[5] == 'f' else (2 if cmd[5] == 'r' else 0)
                             rows.append(f'{indent}{indentOfTryBlock}__b = {cmd}')
+                        elif (cmd.startswith('f"') or cmd.startswith("f'")):
+                            fr = 1
+                            rows.append(f'{indent}{indentOfTryBlock}__b = f""" {cmd[2:-1]} """')
+                        elif (cmd.startswith('r"') or cmd.startswith("r'")):
+                            fr = 2
+                            rows.append(f'{indent}{indentOfTryBlock}__b = r""" {cmd[2:-1]} """')
                         else:
+                            fr = 0
                             rows.append(f'{indent}{indentOfTryBlock}__b = """ {cmd} """')
                         # 前後要留空白，避免 "與"""混在一起
-                        rows.append(f'{indent}{indentOfTryBlock}__c = SSHScriptDollar(None,__b,locals(),globals(),inWith={bool(asPartOfWith)})')
+                        rows.append(f'{indent}{indentOfTryBlock}__c = SSHScriptDollar(None,__b,locals(),globals(),inWith={bool(asPartOfWith)},fr={fr})')
                         rows.append(f'{indent}{indentOfTryBlock}__ret = __c({isTwodollars})')
                         rows.append(f'{indent}finally:')
                         rows.append(f'{indent}{indentOfTryBlock}_c = __c')
@@ -939,11 +946,19 @@ class SSHScript(object):
                         indentOfTryBlock = '    '
                         rows.append(f'{indent}try:')
                         if cmd.startswith('#____'):
+                            fr = 1 if cmd[5] == 'f' else (2 if cmd[5] == 'r' else 0)
                             rows.append(f'{indent}{indentOfTryBlock}__b = {cmd}')
+                        elif (cmd.startswith('f"') or cmd.startswith("f'")):
+                            fr = 1
+                            rows.append(f'{indent}{indentOfTryBlock}__b = f""" {cmd[2:-1]} """')
+                        elif (cmd.startswith('r"') or cmd.startswith("r'")):
+                            fr = 2
+                            rows.append(f'{indent}{indentOfTryBlock}__b = r""" {cmd[2:-1]} """')
                         else:
+                            fr = 0
                             rows.append(f'{indent}{indentOfTryBlock}__b = """ {cmd} """')
                         # 前後要留空白，避免 "與"""混在一起
-                        rows.append(f'{indent}{indentOfTryBlock}__c = SSHScriptDollar(None,__b,locals(),globals(),inWith={bool(asPartOfWith)})')
+                        rows.append(f'{indent}{indentOfTryBlock}__c = SSHScriptDollar(None,__b,locals(),globals(),inWith={bool(asPartOfWith)},fr={fr})')
                         rows.append(f'{indent}{indentOfTryBlock}__ret = __c({isTwodollars})')
                         rows.append(f'{indent}finally:')
                         rows.append(f'{indent}{indentOfTryBlock}_c = __c')
@@ -961,7 +976,7 @@ class SSHScript(object):
 
     
     def run(self,script,varLocals=None,varGlobals=None,showScript=False):
-
+        global newline
         _locals = locals()
         if varLocals is not None:
             _locals.update(varLocals)
@@ -976,24 +991,52 @@ class SSHScript(object):
             rows = self.parseScript(script,_locals=_locals,quotes=quotes)
             if len(rows):
                 # close this block
-                self.blocksOfScript.append('\n'.join(rows))        
+                self.blocksOfScript.append(newline.join(rows))        
         
         assert  self.blocksOfScript is not None
         
-        #def dumpScriptItem(scriptChunk):
-        #    for idx, line in enumerate(scriptChunk.split('\n')):
-        #        print(f'{str(idx+1).zfill(3)}:{line}')
+        # efficiently restore key to content in quotes
+        global skipedKey,replacedKey
+        replacedKey = set()
+        skipedKey = []
+        def iterKey():
+            global skipedKey,replacedKey
+            # reverse key list is faster than using set()
+            allKeys = reversed(quotes.keys())
+            while True:
+                for key in allKeys:
+                    if key in replacedKey: continue
+                    yield key
+                if len(skipedKey) == 0: break
+                allKeys = skipedKey
+                skipedKey = []
+        def pqoute5Sub(m):
+            key = m.group(0)
+            replacedKey.add(key)
+            return quotes[key]
         for scriptChunk in self.blocksOfScript:
             if isinstance(scriptChunk, str):
                 # restore triple-quotes content
-                #for key,(fr,quote,content) in quotes.items():
-                #    c = f'{fr}{quote}{content}{quote[-3:]}'
-                for key,c in quotes.items(): #miso
-                    # 被保留下來的字串中，如果有f-string則eval當中的{$.stdout}
-                    if c.startswith('f"') or c.startswith("f'"):
-                        c = pstdInCurlyBracket.sub(pstdInCurlyBracketSub,c)
-                    scriptChunk = scriptChunk.replace(key,c)
-
+                keygen = iterKey()
+                while True:
+                    try:
+                        key = next(keygen)
+                    except StopIteration:
+                        break
+                    else:
+                        # 被保留下來的字串中，如果有f-string則eval當中的{$.stdout}
+                        if scriptChunk.find(key) == -1:
+                            skipedKey.append(key)
+                        else:
+                            replacedKey.add(key)
+                            c = quotes[key]
+                            if c.startswith('f"') or c.startswith("f'"):
+                                c = pstdInCurlyBracket.sub(pstdInCurlyBracketSub,c)
+                                # 把#4.4#換為雙引號字串
+                                c = pqoute5.sub(pqoute5Sub,c)
+                                # 再掃一次，把{$.stdout}換成_c.stdout
+                                c = pstdInCurlyBracket.sub(pstdInCurlyBracketSub,c)
+                            scriptChunk = scriptChunk.replace(key,c)
                 if showScript:
                     if os.environ.get('DEBUG'):
                         self.lineNumberCount = 0
@@ -1004,7 +1047,6 @@ class SSHScript(object):
                         print(f'{str(self.lineNumberCount).zfill(digits)}:{line}')
                     print()
                     continue
-
                 try:
                     exec(scriptChunk,_globals,_locals)
                 except SyntaxError:
